package ru.dbotthepony.kstarbound.lua

import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonPrimitive
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.json.InternedJsonElementAdapter
import java.io.Closeable
import java.lang.ref.Cleaner.Cleanable
import java.lang.ref.WeakReference

interface LuaHandle : Closeable {
	fun push(into: LuaThread)
	override fun close() {}
	fun toJson(): JsonElement
	val type: LuaType

	object Nil : LuaHandle {
		override fun push(into: LuaThread) {
			into.push()
		}

		override fun toString(): String {
			return "LuaHandle.Nil"
		}

		override fun toJson(): JsonElement {
			return JsonNull.INSTANCE
		}

		override val type: LuaType
			get() = LuaType.NIL
	}

	object True : LuaHandle {
		override fun push(into: LuaThread) {
			into.push(true)
		}

		override fun toString(): String {
			return "LuaHandle.True"
		}

		override fun toJson(): JsonElement {
			return InternedJsonElementAdapter.TRUE
		}

		override val type: LuaType
			get() = LuaType.BOOLEAN
	}

	object False : LuaHandle {
		override fun push(into: LuaThread) {
			into.push(false)
		}

		override fun toString(): String {
			return "LuaHandle.False"
		}

		override fun toJson(): JsonElement {
			return InternedJsonElementAdapter.FALSE
		}

		override val type: LuaType
			get() = LuaType.BOOLEAN
	}

	class LLong(val value: Long) : LuaHandle {
		override fun push(into: LuaThread) {
			into.push(value)
		}

		override fun equals(other: Any?): Boolean {
			return other is LLong && other.value == value
		}

		override fun hashCode(): Int {
			return value.hashCode()
		}

		override fun toString(): String {
			return "LuaHandle.LLong[$value]"
		}

		override fun toJson(): JsonElement {
			return JsonPrimitive(value)
		}

		override val type: LuaType
			get() = LuaType.NUMBER
	}

	class LDouble(val value: Double) : LuaHandle {
		override fun push(into: LuaThread) {
			into.push(value)
		}

		override fun equals(other: Any?): Boolean {
			return other is LDouble && other.value == value
		}

		override fun hashCode(): Int {
			return value.hashCode()
		}

		override fun toString(): String {
			return "LuaHandle.LDouble[$value]"
		}

		override fun toJson(): JsonElement {
			return JsonPrimitive(value)
		}

		override val type: LuaType
			get() = LuaType.NUMBER
	}

	class Regular(private val parent: LuaSharedState, val handle: Int, val key: Any?) : LuaHandle {
		private val cleanable: Cleanable

		var isValid = true
			private set

		override val type: LuaType by lazy(LazyThreadSafetyMode.NONE) {
			check(isValid) { "Tried to use NULL handle!" }
			parent.handlesThread.typeAt(handle)
		}

		init {
			val parent = WeakReference(parent)
			val handle = handle
			val key = key

			cleanable = Starbound.CLEANER.register(this) {
				parent.get()?.freeHandle(handle, key)
			}

		}

		override fun push(into: LuaThread) {
			check(isValid) { "Tried to use NULL handle!" }
			parent.handlesThread.push()
			parent.handlesThread.copy(handle, -1)
			parent.handlesThread.moveStackValuesOnto(into)
		}

		override fun close() {
			if (!isValid) return
			cleanable.clean()
			isValid = false
		}

		override fun toJson(): JsonElement {
			check(isValid) { "Tried to use NULL handle!" }
			return parent.handlesThread.getJson(handle)!!
		}

		override fun toString(): String {
			return "LuaHandle[$parent - $handle / $key]"
		}
	}

	companion object {
		fun boolean(of: Boolean): LuaHandle {
			if (of) return True else return False
		}
	}
}
