package ru.dbotthepony.kstarbound.world.entities

import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import ru.dbotthepony.kommons.gson.contains
import ru.dbotthepony.kommons.gson.get
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.AssetPath
import ru.dbotthepony.kstarbound.defs.EntityType
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.json.jsonArrayOf
import ru.dbotthepony.kstarbound.json.putAll
import ru.dbotthepony.kstarbound.json.readJsonElement
import ru.dbotthepony.kstarbound.json.writeJsonElement
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.LuaUpdateComponent
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.network.syncher.networkedFloat
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.entities.api.ScriptedEntity
import java.io.DataInputStream
import java.io.DataOutputStream
import kotlin.properties.Delegates

class StagehandEntity(isRemote: Boolean = false) : AbstractEntity(), ScriptedEntity {
	constructor(stream: DataInputStream, isLegacy: Boolean) : this(stream.readJsonElement().asJsonObject, true)

	constructor(data: JsonObject, isRemote: Boolean = false) : this(isRemote) {
		deserialize(data)
	}

	constructor(type: String, overrides: JsonElement? = JsonNull.INSTANCE) : this(Registries.makeStagehandConfig(type, overrides))

	var xPosition by networkedFloat().also { networkGroup.upstream.add(it) }
	var yPosition by networkedFloat().also { networkGroup.upstream.add(it) }

	init {
		networkGroup.upstream.add(uniqueID)
		this.isRemote = isRemote
	}

	override var position: Vector2d
		get() = Vector2d(xPosition, yPosition)
		set(value) {
			xPosition = value.x
			yPosition = value.y
		}

	override val type: EntityType
		get() = EntityType.STAGEHAND

	private var config: JsonObject = JsonObject()

	val typeName: String
		get() = config["type"]?.asString ?: ""

	var isScripted = false
		private set

	override var lua by Delegates.notNull<LuaThread>()
		private set

	var luaUpdate by Delegates.notNull<LuaUpdateComponent>()
		private set

	override fun deserialize(data: JsonObject) {
		super.deserialize(data)

		config = data.deepCopy()
		isScripted = config["scripts"] is JsonArray

		if ("position" in config) {
			val pos = config["position"].asJsonArray
			xPosition = pos[0].asDouble
			yPosition = pos[1].asDouble
		}

		var broadcastArea: AABB? = null

		if ("broadcastArea" in config) {
			broadcastArea = Starbound.gson.fromJsonFast(config["broadcastArena"], AABB::class.java)

			if (broadcastArea.width < 0.0 || broadcastArea.height > 0.0)
				broadcastArea = null
		}

		boundingBox = broadcastArea ?: AABB.withSide(Vector2d.ZERO, 5.0)
	}

	override fun onJoinWorld(world: World<*, *>) {
		super.onJoinWorld(world)

		if (isLocal && isScripted) {
			lua = LuaThread()
			luaUpdate = LuaUpdateComponent(lua, this)
			lua.attach(config["scripts"].asJsonArray.map { AssetPath(it.asString) })
			luaUpdate.stepCount = config.get("scriptDelta", 5.0)

			if ("scriptStorage" in config) {
				lua.push(config["scriptStorage"])
				lua.storeGlobal("storage")
			} else {
				lua.pushTable()
				lua.storeGlobal("storage")
			}

			lua.initScripts()
		}
	}

	override fun uninit(world: World<*, *>) {
		if (isLocal && isScripted) {
			lua.close()
		}
	}

	override fun serialize(data: JsonObject) {
		super.serialize(data)

		data.putAll(config, true)
		data["position"] = jsonArrayOf(JsonPrimitive(xPosition), JsonPrimitive(yPosition))

		if (uniqueID.get() != null)
			data["uniqueId"] = uniqueID.get()!!
		else
			data.remove("uniqueId")

		if (isScripted) {
			lua.loadGlobal("storage")
			data["scriptStorage"] = lua.popJson()
		}
	}

	override fun writeNetwork(stream: DataOutputStream, isLegacy: Boolean) {
		stream.writeJsonElement(config)
	}

	private var boundingBox = AABB.withSide(Vector2d.ZERO, 5.0)

	override val metaBoundingBox: AABB
		get() = boundingBox + position
	override val name: String
		get() = "Stagehand"
	override val description: String
		get() = "Technical entity responsible for doing cool stuff"

	override fun callScript(fnName: String, arguments: JsonArray): JsonElement {
		require(isLocal) { "Calling script on remote entity" }
		return super.callScript(fnName, arguments)
	}

	override fun evalScript(code: String): JsonElement {
		require(isLocal) { "Calling script on remote entity" }
		return super.evalScript(code)
	}
}
